package org.obolibrary.robot;

import java.io.*;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.stream.Collectors;
import org.apache.commons.io.FileUtils;
import org.apache.jena.query.*;
import org.apache.jena.tdb.TDBFactory;
import org.apache.jena.tdb.transaction.TDBTransactionException;
import org.apache.poi.ss.usermodel.Workbook;
import org.obolibrary.robot.checks.Report;
import org.obolibrary.robot.checks.ReportQuery;
import org.obolibrary.robot.checks.Violation;
import org.obolibrary.robot.export.Table;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Report issues with an ontology using a series of QC SPARQL queries.
 *
 * @author <a href="mailto:rbca.jackson@gmail.com">Rebecca Jackson</a>
 * @author <a href="mailto:james@overton.ca">James A. Overton</a>
 */
public class ReportOperation {

  /** Directory for queries. */
  private static final String queryDir = "report_queries";

  /** Logger. */
  private static final Logger logger = LoggerFactory.getLogger(ReportOperation.class);

  /** Namespace for general input error messages. */
  private static final String NS = "report#";

  /** Error message when user profiles an invalid fail-on level. */
  private static final String failOnError = NS + "FAIL ON ERROR '%s' is not a valid fail-on level.";

  /** Error message when 'limit' is not a number. */
  private static final String limitNumberError =
      NS + "LIMIT NUMBER ERROR --limit argument '%s' must be an integer.";

  /** Error message when the query does not have ?entity. */
  private static final String missingEntityBinding =
      NS + "MISSING ENTITY BINDING query '%s' must include an '?entity'";

  /** Error message when a user-provided report query does not exist. */
  private static final String missingQueryError =
      NS + "MISSING QUERY ERROR query at '%s' does not exist.";

  /** Error message when 'print' is not a number. */
  private static final String printNumberError =
      NS + "PRINT NUMBER ERROR --print argument '%s' must be an integer.";

  /** Error message when user provides a rule level other than INFO, WARN, or ERROR. */
  private static final String reportLevelError =
      NS + "REPORT LEVEL ERROR '%s' is not a valid reporting level.";

  /** Error message when a report query name or path in report profile cannot be found. */
  private static final String unknownReportQuery =
      NS + "UNKNOWN REPORT QUERY one or more rule names ('%s') are not valid default rules";

  /** Reporting level INFO. */
  private static final String INFO = "INFO";

  /** Reporting level WARN. */
  private static final String WARN = "WARN";

  /** Reporting level ERROR. */
  private static final String ERROR = "ERROR";

  /**
   * Return a map from option name to default option value, for all the available report options.
   *
   * @return a map with default values for all available options
   */
  public static Map<String, String> getDefaultOptions() {
    Map<String, String> options = new HashMap<>();
    options.put("print", "0");
    options.put("fail-on", "error");
    options.put("labels", "false");
    options.put("format", null);
    options.put("profile", null);
    options.put("base-iri", null);
    options.put("tdb", "false");
    options.put("tdb-directory", ".tdb");
    options.put("keep-tdb-mappings", "false");
    options.put("standalone", "true");
    return options;
  }

  /**
   * Report on the ontology using the rules within the profile and print results. Prefer
   * report(OWLOntology ontology, String profilePath, String outputPath, String format, String
   * failOn).
   *
   * @param ontology the OWLOntology to report
   * @param ioHelper IOHelper to work with ontology
   * @throws Exception on any reporting error
   */
  public static void report(OWLOntology ontology, IOHelper ioHelper) throws Exception {
    report(ontology, ioHelper, null, null, null, null);
  }

  /**
   * Given an ontology, a profile path (or null), an output path (or null), and a report format (or
   * null) report on the ontology using the rules within the profile and write results to the output
   * path. If profile is null, use the default profile in resources. If the output path is null,
   * write results to console. If the format is null, write results in TSV format.
   *
   * @param ontology OWLOntology to report on
   * @param profilePath user profile file path to use, or null
   * @param outputPath string path to write report file to, or null
   * @param format string format for the output report (TSV or YAML), or null
   * @param failOn logging level to fail execution
   * @return true if successful, false if failed
   * @throws Exception on any error
   */
  public static boolean report(
      OWLOntology ontology, String profilePath, String outputPath, String format, String failOn)
      throws Exception {
    return report(ontology, null, profilePath, outputPath, format, failOn, false);
  }

  /**
   * Given an ontology, an IOHelper, a profile path (or null), an output path (or null), a report
   * format (or null), and a level to fail on, report on the ontology using the rules within the
   * profile and write results to the output path. If profile is null, use the default profile in
   * resources. If the output path is null, write results to console. If the format is null, write
   * results in TSV format. Exit with status 1 if any violations of fail-on level are found.
   *
   * @param ontology OWLOntology to report on
   * @param ioHelper IOHelper to use
   * @param profilePath user profile file path to use, or null
   * @param outputPath string path to write report file to, or null
   * @param format string format for the output report (TSV or YAML), or null
   * @param failOn logging level to fail execution
   * @return true if successful, false if failed
   * @throws Exception on any error
   */
  public static boolean report(
      OWLOntology ontology,
      IOHelper ioHelper,
      String profilePath,
      String outputPath,
      String format,
      String failOn)
      throws Exception {
    return report(ontology, ioHelper, profilePath, outputPath, format, failOn, false);
  }

  /**
   * Report on the ontology using the rules within the profile and print results. Prefer
   * report(OWLOntology ontology, IOHelper ioHelper, String outputPath, Map&lt;String,String&gt;
   * options).
   *
   * @param ontology the OWLOntology to report
   * @param ioHelper IOHelper to work with ontology
   * @param options map of report options
   * @throws Exception on any reporting error
   */
  public static void report(OWLOntology ontology, IOHelper ioHelper, Map<String, String> options)
      throws Exception {
    report(ontology, ioHelper, null, options);
  }

  /**
   * Given an ontology, an IOHelper, a profile path (or null), an output path (or null), a report
   * format (or null), a level to fail on, and a boolean indicating to use labels, report on the
   * ontology using the rules within the profile and write results to the output path. If profile is
   * null, use the default profile in resources. If the output path is null, write results to
   * console. If the format is null, write results in TSV format. Exit with status 1 if any
   * violations of fail-on level are found.
   *
   * @param ontology OWLOntology to report on
   * @param ioHelper IOHelper to use
   * @param profilePath user profile file path to use, or null
   * @param outputPath string path to write report file to, or null
   * @param format string format for the output report (TSV or YAML), or null
   * @param failOn logging level to fail execution
   * @param useLabels if true, use labels for output
   * @return true if successful, false if failed
   * @throws Exception on any error
   */
  public static boolean report(
      OWLOntology ontology,
      IOHelper ioHelper,
      String profilePath,
      String outputPath,
      String format,
      String failOn,
      boolean useLabels)
      throws Exception {
    Map<String, String> options = getDefaultOptions();
    if (profilePath != null) {
      options.put("profile", profilePath);
    }
    if (format != null) {
      options.put("format", format);
    }
    if (failOn != null) {
      options.put("fail-on", failOn);
    }
    if (useLabels) {
      options.put("labels", "true");
    }
    return report(ontology, ioHelper, outputPath, options);
  }

  /**
   * Given an ontology, an IOHelper, an output path (or null), and a map of options (or null),
   * report on the ontology using the rules within the profile specified by the options and write
   * results to the output path. If profile is null, use the default profile in resources. If the
   * output path is null, write results to console. If the format is null, write results in TSV
   * format.
   *
   * @param ontology the OWLOntology to report
   * @param ioHelper IOHelper to work with ontology
   * @param outputPath string path to write report file to, or null
   * @param options map of report options
   * @return false if there are violations at or above the fail-on level, true otherwise
   * @throws Exception on any reporting error
   */
  public static boolean report(
      OWLOntology ontology, IOHelper ioHelper, String outputPath, Map<String, String> options)
      throws Exception {
    // Generate the report object with violation details
    Report report = getReport(ontology, ioHelper, options);
    return processReport(report, outputPath, options);
  }

  /**
   * Given an ontology, an IOHelper, and a map of options, create a Report object and run the report
   * queries specified in a profile (from options, or default). Return the completed Report object.
   *
   * @param ontology OWLOntology to report on
   * @param ioHelper IOHelper to resolve labels
   * @param options Map of report options
   * @return Report object with violation details
   * @throws Exception on any reporting error
   */
  public static Report getReport(
      OWLOntology ontology, IOHelper ioHelper, Map<String, String> options) throws Exception {
    // Get options specified in map or default options
    if (options == null) {
      options = getDefaultOptions();
    }

    String profilePath = OptionsHelper.getOption(options, "profile", null);
    boolean useLabels = OptionsHelper.optionIsTrue(options, "labels");

    // The profile is a map of rule name and reporting level
    Map<String, String> profile = getProfile(profilePath);
    // We get the queries for the rules from the profile
    Set<ReportQuery> reportQueries = getReportQueries(profile);

    Dataset dataset = QueryOperation.loadOntologyAsDataset(ontology, false);

    // Create the report object
    Report report;
    if (ioHelper != null) {
      report = new Report(ontology, ioHelper, useLabels);
    } else {
      report = new Report(ontology, useLabels);
    }

    for (ReportQuery rq : reportQueries) {
      String queryName = rq.getRuleName();
      String fullQueryString = rq.getQuery();
      String queryString;
      // Remove any comments
      List<String> lines = new ArrayList<>();
      for (String line : fullQueryString.split("\n")) {
        if (!line.startsWith("#")) {
          lines.add(line);
        }
      }
      queryString = String.join("\n", lines);
      // Use the query to get violations
      List<Violation> violations =
          getViolations(ioHelper, dataset, queryName, queryString, options);
      // If violations is not returned properly, the query did not have the correct format
      if (violations == null) {
        throw new Exception(String.format(missingEntityBinding, queryName));
      }
      rq.addViolations(violations);
      report.addReportQuery(rq);
    }
    return report;
  }

  /**
   * Given an input path, an output path (or null), and a map of options, report on the ontology
   * using the rules within the profile specified by the options and write results to the output
   * path. Ontology is loaded to dataset backed on disk. The labels option is not supported with
   * TDB.
   *
   * @param inputPath String path of ontology to load
   * @param outputPath String path to write report file to, or null
   * @param options map of report options
   * @return false if there are violations at or above the fail-on level, true otherwise
   * @throws Exception on any reporting error
   */
  public static boolean tdbReport(String inputPath, String outputPath, Map<String, String> options)
      throws Exception {
    Report report = getTDBReport(inputPath, options);
    return processReport(report, outputPath, options);
  }

  /**
   * Given an input path to an ontology and a map of options, create a Report object and run (on TDB
   * dataset) the report queries specified in a profile (from options, or default). Return the
   * completed Report object. The labels option is not supported with TDB.
   *
   * @param inputPath path to load triples to TDB
   * @param options map of report options
   * @return Report object with violation details
   * @throws Exception on any loading, query, or reporting error
   */
  public static Report getTDBReport(String inputPath, Map<String, String> options)
      throws Exception {
    String tdbDir = OptionsHelper.getOption(options, "tdb-directory", ".tdb");

    // Load dataset
    // Fail if the input path is not in RDF/XML or TTL
    Dataset dataset = IOHelper.loadToTDBDataset(inputPath, tdbDir);

    Report report;
    boolean keepMappings = OptionsHelper.optionIsTrue(options, "keep-tdb-mappings");
    try {
      report = getTDBReport(dataset, options);
    } finally {
      // Close and release
      dataset.close();
      try {
        TDBFactory.release(dataset);
      } catch (TDBTransactionException e) {
        // Do nothing - already released
      }

      if (!keepMappings) {
        // Maybe delete
        boolean success = IOHelper.cleanTDB(tdbDir);
        if (!success) {
          logger.error(String.format("Unable to remove directory '%s'", tdbDir));
        }
      }
    }

    return report;
  }

  /**
   * Given a dataset and a map of options, create a Report object and run (on TDB dataset) the
   * report queries specified in a profile (from options, or default). Return the completed Report
   * object. The labels option is not supported with TDB.
   *
   * @param dataset TDB Dataset to perform Report operation on
   * @param options Map of report options
   * @return Report object with violation details
   * @throws Exception on any reporting error
   */
  public static Report getTDBReport(Dataset dataset, Map<String, String> options) throws Exception {
    IOHelper ioHelper = new IOHelper();
    return getTDBReport(ioHelper, dataset, options);
  }

  /**
   * Given a dataset and a map of options, create a Report object and run (on TDB dataset) the
   * report queries specified in a profile (from options, or default). Return the completed Report
   * object. The labels option is not supported with TDB.
   *
   * @param ioHelper IOHelper to resolve IRIs
   * @param dataset TDB Dataset to perform Report operation on
   * @param options Map of report options
   * @return Report object with violation details
   * @throws Exception on any reporting error
   */
  public static Report getTDBReport(IOHelper ioHelper, Dataset dataset, Map<String, String> options)
      throws Exception {
    // Get options specified in map or default options
    if (options == null) {
      options = getDefaultOptions();
    }
    // Make sure TDB is true here
    options.put("tdb", "true");

    String profilePath = OptionsHelper.getOption(options, "profile", null);

    // The profile is a map of rule name and reporting level
    Map<String, String> profile = getProfile(profilePath);
    // We get the queries for the rules from the profile
    Set<ReportQuery> reportQueries = getReportQueries(profile);

    boolean useLabels = OptionsHelper.optionIsTrue(options, "labels");
    Map<IRI, String> labelMap = null;
    if (useLabels) {
      labelMap = new HashMap<>();
      // Run query over dataset to retrive all labels
      String query =
          "SELECT ?s ?label WHERE { ?s <http://www.w3.org/2000/01/rdf-schema#label> ?label }";
      dataset.begin(ReadWrite.READ);
      try {
        ResultSet labelResults = QueryOperation.execQuery(dataset, query);
        while (labelResults.hasNext()) {
          QuerySolution qs = labelResults.next();
          IRI iri = IRI.create(qs.getResource("s").getURI());
          String label = qs.getLiteral("label").getString();
          labelMap.put(iri, label);
        }
      } finally {
        dataset.end();
      }
    }

    // Create the report object (maybe using labels)
    Report report = new Report(labelMap);

    for (ReportQuery rq : reportQueries) {
      String queryName = rq.getRuleName();
      String fullQueryString = rq.getQuery();
      String queryString;
      // Remove any comments
      List<String> lines = new ArrayList<>();
      for (String line : fullQueryString.split("\n")) {
        if (!line.startsWith("#")) {
          lines.add(line);
        }
      }
      queryString = String.join("\n", lines);
      // Use the query to get violations
      List<Violation> violations =
          getViolations(ioHelper, dataset, queryName, queryString, options);
      // If violations is not returned properly, the query did not have the correct format
      if (violations == null) {
        throw new Exception(String.format(missingEntityBinding, queryName));
      }
      rq.addViolations(violations);
      report.addReportQuery(rq);
    }
    return report;
  }

  /**
   * Given a Report, an output path, and a map of report options, process the Report results and
   * save the report to the output path.
   *
   * @param report completed Report object
   * @param outputPath path to save report to
   * @param options Map of report options
   * @return true if report passed, false if it failed
   * @throws IOException on issue writing to file
   */
  public static boolean processReport(Report report, String outputPath, Map<String, String> options)
      throws IOException {
    // Print violations to terminal
    int violationCount = report.getTotalViolations();
    if (violationCount != 0) {
      System.out.println("Violations: " + violationCount);
      System.out.println("-----------------");
      System.out.println(ERROR + ":      " + report.getTotalViolations(ERROR));
      System.out.println(WARN + ":       " + report.getTotalViolations(WARN));
      System.out.println(INFO + ":       " + report.getTotalViolations(INFO));
    } else {
      System.out.println("No violations found.");
    }

    // Maybe print some of the lines
    String printString = OptionsHelper.getOption(options, "print", "0").trim();
    // Parse print N lines option to an int
    int print;
    try {
      print = Integer.parseInt(printString);
    } catch (NumberFormatException e) {
      // Not a number
      throw new IllegalArgumentException(String.format(printNumberError, printString));
    }

    // Format is determined either by --format or the extension of the output path
    String format = OptionsHelper.getOption(options, "format");
    if (format == null && outputPath != null) {
      format = outputPath.substring(outputPath.lastIndexOf(".") + 1);
    } else if (format == null) {
      // Null format means no output file, will be printed as TSV
      format = "tsv";
    }

    // Process different output formats while writing print lines if requested
    // First check if format is JSON or YAML
    // We don't use the Table for these formats because we want to group by violation level and rule
    // name
    if (format.equalsIgnoreCase("yaml")) {
      if (print > 0) {
        // We use the Table to get a list of rows to print as TSV
        Table t = report.toTable("tsv");
        printNViolations(t.toList(""), print, "\t");
      }
      String yaml = report.toYAML();
      try (FileWriter fw = new FileWriter(outputPath);
          BufferedWriter bw = new BufferedWriter(fw)) {
        logger.debug("Writing report to: " + outputPath);
        bw.write(yaml);
      }

    } else if (format.equalsIgnoreCase("json")) {
      if (print > 0) {
        // We use the Table to get a list of rows to print as TSV
        Table t = report.toTable("tsv");
        printNViolations(t.toList(""), print, "\t");
      }
      String json = report.toJSON();
      try (FileWriter fw = new FileWriter(outputPath);
          BufferedWriter bw = new BufferedWriter(fw)) {
        logger.debug("Writing report to: " + outputPath);
        bw.write(json);
      }

    } else {
      // All other formats are converted to export Table to get the output
      Table reportTable = report.toTable(format);
      List<String[]> rows;
      switch (format.toLowerCase()) {
        case "html":
          if (print > 0) {
            printNViolations(reportTable.toList(""), print, "\t");
          }
          String html = reportTable.toHTML("", OptionsHelper.optionIsTrue(options, "standalone"));
          try (FileWriter fw = new FileWriter(outputPath);
              BufferedWriter bw = new BufferedWriter(fw)) {
            logger.debug("Writing report to: " + outputPath);
            bw.write(html);
          }
          break;

        case "csv":
          rows = reportTable.toList("");
          if (outputPath != null) {
            if (print > 0) {
              printNViolations(rows, print, ",");
            }
            try (FileWriter fw = new FileWriter(outputPath);
                BufferedWriter bw = new BufferedWriter(fw)) {
              logger.debug("Writing report to: " + outputPath);
              IOHelper.writeTable(rows, bw, ',');
            }
          } else {
            if (print == 0) {
              print = rows.size();
            }
            printNViolations(rows, print, ",");
          }
          break;

        case "xlsx":
          if (print > 0) {
            printNViolations(reportTable.toList(""), print, "\t");
          }
          try (Workbook wb = reportTable.asWorkbook("");
              FileOutputStream fos = new FileOutputStream(outputPath)) {
            wb.write(fos);
          }
          break;

        case "tsv":
        default:
          rows = reportTable.toList("");
          if (outputPath != null) {
            if (print > 0) {
              printNViolations(rows, print, "\t");
            }
            try (FileWriter fw = new FileWriter(outputPath);
                BufferedWriter bw = new BufferedWriter(fw)) {
              logger.debug("Writing report to: " + outputPath);
              IOHelper.writeTable(rows, bw, '\t');
            }
          } else {
            if (print == 0) {
              print = rows.size();
            }
            printNViolations(rows, print, "\t");
          }
      }
    }

    String failOn = OptionsHelper.getOption(options, "fail-on", "error");
    // Set failOn if null to default
    if (failOn == null) {
      failOn = ERROR;
    }

    // If a fail-on is provided, return false if there are violations of the given level
    if (failOn.equalsIgnoreCase("none")) {
      return true;
    } else if (failOn.equalsIgnoreCase(ERROR)) {
      return report.getTotalViolations(ERROR) <= 0;
    } else if (failOn.equalsIgnoreCase(WARN)) {
      return (report.getTotalViolations(ERROR) + report.getTotalViolations(WARN)) <= 0;
    } else if (failOn.equalsIgnoreCase(INFO)) {
      return report.getTotalViolations() <= 0;
    } else {
      throw new IllegalArgumentException(String.format(failOnError, failOn));
    }
  }

  /**
   * Create a list of ReportQuery objects based on the profile.
   *
   * @param profile Map of rule name to violation level
   * @return List of Report Query objects
   * @throws IOException on problem getting query strings
   * @throws URISyntaxException on problem getting query strings
   */
  private static Set<ReportQuery> getReportQueries(Map<String, String> profile)
      throws IOException, URISyntaxException {
    Set<ReportQuery> reportQueries = new HashSet<>();
    Set<String> defaultRules = new HashSet<>();
    Set<String> userRules = new HashSet<>();
    for (String rule : profile.keySet()) {
      if (rule.startsWith("file:")) {
        userRules.add(rule);
      } else {
        defaultRules.add(rule);
      }
    }

    Map<String, String> defaultQueryStrings = getDefaultQueryStrings(defaultRules);
    for (Map.Entry<String, String> qs : defaultQueryStrings.entrySet()) {
      String ruleName = qs.getKey();
      String ruleURL = "http://robot.obolibrary.org/report_queries/" + ruleName;
      ReportQuery rq = new ReportQuery(ruleName, ruleURL, qs.getValue(), profile.get(ruleName));
      reportQueries.add(rq);
    }
    Map<String, String> userQueryStrings = getUserQueryStrings(userRules);
    for (Map.Entry<String, String> qs : userQueryStrings.entrySet()) {
      String rulePath = qs.getKey();
      String ruleName =
          rulePath.substring(rulePath.lastIndexOf("/") + 1, rulePath.lastIndexOf("."));
      ReportQuery rq = new ReportQuery(ruleName, qs.getValue(), profile.get(rulePath));
      reportQueries.add(rq);
    }
    return reportQueries;
  }

  /**
   * Given a set of user-provided query paths for a set of rules, return a map of the rule names and
   * the file (query) contents.
   *
   * @param rules set of file paths to user query files
   * @return map of rule name and query string
   * @throws URISyntaxException on issue converting file path URL to URI
   * @throws IOException on any issue reading the file
   */
  private static Map<String, String> getUserQueryStrings(Set<String> rules)
      throws URISyntaxException, IOException {
    Map<String, String> queries = new HashMap<>();
    for (String rule : rules) {
      if (rule.startsWith("file:///")) {
        // Process an absolute path
        File file = new File(new URL(rule).toURI());
        if (!file.exists()) {
          throw new IOException(String.format(missingQueryError, file.getPath()));
        }
        queries.put(rule, FileUtils.readFileToString(file, Charset.defaultCharset()));
      } else {
        // Process a relative path
        String path = rule.substring(5);
        File file = new File(path);
        if (!file.exists()) {
          throw new IOException(String.format(missingQueryError, file.getPath()));
        }
        queries.put(rule, FileUtils.readFileToString(file, Charset.defaultCharset()));
      }
    }
    return queries;
  }

  /**
   * Given a set of default rules, return a map of the rule names and query strings. This is a
   * "brute-force" method to retrieve default query file contents from packaged jar.
   *
   * @param rules subset of the default rules to include
   * @return map of rule name and query string
   * @throws URISyntaxException on issue converting path to URI
   * @throws IOException on any issue with accessing files or file contents
   */
  private static Map<String, String> getDefaultQueryStrings(Set<String> rules)
      throws IOException, URISyntaxException {
    URL dirURL = ReportOperation.class.getClassLoader().getResource(queryDir);
    Map<String, String> queries = new HashMap<>();
    // Handle simple file path, probably accessed during testing
    if (dirURL != null && dirURL.getProtocol().equals("file")) {
      final Set<Path> queryFilePaths =
          Files.list(Paths.get(dirURL.toURI())).collect(Collectors.toSet());
      if (queryFilePaths.size() == 0) {
        throw new IOException(
            "Cannot access report query files. There are no files in the directory.");
      }
      Set<String> defaultRuleNames = new HashSet<>();
      for (final Path qPath : queryFilePaths) {
        final String ruleName = qPath.getFileName().toString().split("[.]")[0];
        defaultRuleNames.add(ruleName);
        // Only add it to the queries if the rule set contains that rule
        // If rules == null, include all rules
        if (rules == null || rules.contains(ruleName)) {
          queries.put(
              ruleName, FileUtils.readFileToString(qPath.toFile(), Charset.defaultCharset()));
        }
      }
      // Check that all the provided rule names are valid defaults
      if (rules != null) {
        rules.removeAll(defaultRuleNames);
        if (rules.size() > 0) {
          throw new IOException(String.format(unknownReportQuery, String.join(", ", rules)));
        }
      }
      return queries;
    }
    // Handle inside jar file
    // This will be the case any time someone runs ROBOT via CLI
    if (dirURL == null) {
      String cls = ReportOperation.class.getName().replace(".", "/") + ".class";
      dirURL = ReportOperation.class.getClassLoader().getResource(cls);
    }
    if (dirURL == null) {
      throw new IOException(
          "Cannot access report query files in JAR. The resource does not exist.");
    }
    String protocol = dirURL.getProtocol();
    if (protocol.equals("jar")) {
      String jarPath = dirURL.getPath().substring(5, dirURL.getPath().indexOf("!"));
      // Get all entries in jar
      Enumeration<JarEntry> entries;
      try (JarFile jar = new JarFile(URLDecoder.decode(jarPath, "UTF-8"))) {
        entries = jar.entries();
        if (!entries.hasMoreElements()) {
          throw new IOException(
              "Cannot access report query files in JAR. There are no entries in the JAR.");
        }
        // Track rules that have successfully been retrieved
        Set<String> defaultRuleNames = new HashSet<>();
        while (entries.hasMoreElements()) {
          JarEntry resource = entries.nextElement();
          String resourceName = resource.getName();
          if (resourceName.startsWith(queryDir) && !resourceName.endsWith("/")) {
            // Get just the rule name
            String ruleName =
                resourceName.substring(
                    resourceName.lastIndexOf("/") + 1, resourceName.indexOf(".rq"));
            defaultRuleNames.add(ruleName);
            // Only add it to the queries if the rule set contains that rule
            // If rules == null, include all rules
            if (rules == null || rules.contains(ruleName)) {
              InputStream is = jar.getInputStream(resource);
              StringBuilder sb = new StringBuilder();
              try (BufferedReader br = new BufferedReader(new InputStreamReader(is))) {
                String chr;
                while ((chr = br.readLine()) != null) {
                  sb.append(chr);
                }
              }
              // Remove the headers
              String fullQueryString = sb.toString();
              String queryString = fullQueryString.substring(fullQueryString.indexOf("PREFIX"));
              queries.put(ruleName, queryString);
            }
          }
        }
        // Check that all the provided rule names are valid defaults
        if (rules != null) {
          rules.removeAll(defaultRuleNames);
          if (rules.size() > 0) {
            throw new IOException(String.format(unknownReportQuery, String.join(", ", rules)));
          }
        }
      }
      return queries;
    }
    // If nothing has been returned, it's an exception
    throw new IOException("Cannot access report query files.");
  }

  /**
   * Given the path to a profile file (or null), return the rules and their levels (info, warn, or
   * error). If profile == null, return the default profile in the resources.
   *
   * @param path path to profile, or null
   * @return map of rule name and its reporting level
   * @throws IOException on any issue reading the profile file
   */
  private static Map<String, String> getProfile(String path) throws IOException {
    Map<String, String> profile = new HashMap<>();
    InputStream is;
    // If the file was not provided, get the default
    if (path == null) {
      is = ReportOperation.class.getResourceAsStream("/report_profile.txt");
    } else {
      is = new FileInputStream(path);
    }
    try (BufferedReader br = new BufferedReader(new InputStreamReader(is))) {
      String line;
      while ((line = br.readLine()) != null) {
        String[] split = line.trim().split("\t");
        String level = split[0].toUpperCase();
        // The level should be: INFO, WARN, or ERROR
        if (!INFO.equals(level) && !WARN.equals(level) && !ERROR.equals(level)) {
          throw new IllegalArgumentException(String.format(reportLevelError, split[0]));
        }
        String rule = split[1];
        profile.put(rule, level);
      }
    }
    return profile;
  }

  /**
   * Given a QuerySolution and a String var to retrieve, return the value of the the var. If the var
   * is not in the solution, return null.
   *
   * @param qs QuerySolution
   * @param var String variable to retrieve from result
   * @return string or null
   */
  private static String getQueryResultOrNull(QuerySolution qs, String var) {
    try {
      return qs.get(var).toString();
    } catch (NullPointerException e) {
      return null;
    }
  }

  /**
   * Given an ontology as a Dataset and a query, return the violations found by that query.
   *
   * @param ioHelper with namespaces, for creating IRIs
   * @param dataset the ontology/ontologies as a dataset
   * @param queryName name of query that produced result set
   * @param query the query
   * @param options map of additional option strings
   * @return List of Violations
   * @throws IOException on issue parsing query
   */
  public static List<Violation> getViolations(
      IOHelper ioHelper,
      Dataset dataset,
      String queryName,
      String query,
      Map<String, String> options)
      throws Exception {
    boolean tdb = OptionsHelper.optionIsTrue(options, "tdb");
    String limitString = OptionsHelper.getOption(options, "limit", null);
    Integer limit = null;
    if (limitString != null) {
      try {
        limit = Integer.parseInt(limitString);
      } catch (NumberFormatException e) {
        throw new Exception(String.format(limitNumberError, limitString));
      }
    }

    if (tdb) {
      // If using TDB we must be in a read transaction to query
      dataset.begin(ReadWrite.READ);
      try {
        ResultSet violationSet = QueryOperation.execQuery(dataset, query);
        return getViolationsFromResults(ioHelper, queryName, violationSet, limit);
      } catch (Exception e) {
        // If query fails, return null
        // And warn that report may be incomplete
        logger.error(
            String.format(
                "Could not complete query '%s' - report may be incomplete.\nCause:\n%s",
                queryName, e.getMessage()));
        return null;
      } finally {
        // Always end the transaction
        dataset.end();
      }
    } else {
      try {
        ResultSet violationSet = QueryOperation.execQuery(dataset, query);
        return getViolationsFromResults(ioHelper, queryName, violationSet, limit);
      } catch (Exception e) {
        // If query fails, return null
        // And warn that report may be incomplete
        logger.error(
            String.format(
                "Could not complete query '%s' - report may be incomplete.\nCause:\n%s",
                queryName, e.getMessage()));
        return null;
      }
    }
  }

  private static final OWLDataFactory dataFactory = OWLManager.getOWLDataFactory();

  /**
   * Given a query name, a result set, and a limit for results, return a list of Violation objects
   * for those results.
   *
   * @param ioHelper with namespaces, for creating IRIs
   * @param queryName name of query that produced result set
   * @param violationSet ResultSet of query results
   * @param limit number of results to limit, or null for no limit
   * @return list of Violation objects
   * @throws Exception on malformed query
   */
  private static List<Violation> getViolationsFromResults(
      IOHelper ioHelper, String queryName, ResultSet violationSet, Integer limit) throws Exception {
    List<Violation> violations = new ArrayList<>();

    // Counter for stopping at limit
    int c = 0;
    while (violationSet.hasNext()) {
      if (limit != null && limit <= c) {
        // Stop checking violations
        break;
      }

      // Wrap in try catch in case GC overflows
      // Will return results up to this point with warning that report may be incomplete
      QuerySolution qs;
      try {
        qs = violationSet.next();
      } catch (Exception e) {
        logger.error(
            String.format(
                "Could not retrieve all results for query '%s' - report may be incomplete.\nCause:\n%s",
                queryName, e.getMessage()));
        return violations;
      }

      // entity should never be null (missing entity binding error)
      String entity = getQueryResultOrNull(qs, "entity");
      if (entity == null) {
        throw new Exception(String.format(missingEntityBinding, queryName));
      }

      // skip RDFS and OWL terms
      if (entity.contains("/rdf-schema#") || entity.contains("/owl#")) {
        continue;
      }

      if (!ioHelper.getBaseNamespaces().isEmpty()) {
        boolean base = false;
        for (String base_prefix : ioHelper.getBaseNamespaces()) {
          if (entity.startsWith(base_prefix)) {
            logger.info(
                String.format(
                    "Skipping entity '%s' from report because it follows the base-namespace.",
                    entity));
            base = true;
            break;
          }
        }
        if (!base) {
          continue;
        }
      }

      Violation violation;
      try {
        OWLClass cls = dataFactory.getOWLClass(ioHelper.createIRI(entity));
        violation = new Violation(cls);
      } catch (Exception e) {
        // Blank node, use the string bnode ID
        violation = new Violation("blank node");
      }

      // try and get a property and value from the query
      String property = getQueryResultOrNull(qs, "property");
      String value = getQueryResultOrNull(qs, "value");
      // add details to Violation
      if (property != null) {
        OWLEntity e = dataFactory.getOWLClass(ioHelper.createIRI(property));
        if (value != null) {
          IRI valIRI = ioHelper.createIRI(value);
          if (valIRI != null) {
            OWLEntity v = dataFactory.getOWLClass(valIRI);
            violation.addStatement(e, v);
          } else {
            violation.addStatement(e, value);
          }
        } else {
          violation.addStatement(e, "");
        }
      }
      violations.add(violation);

      // Increase counter
      c++;
    }
    return violations;
  }

  /**
   * Given an array of lines and a number of lines to print, print that number of violations (one
   * per line).
   *
   * @param lines array of lines to print
   * @param n number of lines to print
   * @param separator column separator
   */
  private static void printNViolations(List<String[]> lines, int n, String separator) {
    if (lines.size() <= n) {
      n = lines.size() - 1;
    }
    System.out.println(String.format("\nFirst %d violations:", n));
    for (int i = 0; i < n; i++) {
      // i + 1 to skip headers
      System.out.println(String.join(separator, lines.get(i + 1)));
    }
  }
}
