package org.obolibrary.robot.export;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.dataformat.yaml.YAMLMapper;
import com.google.common.collect.Sets;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import java.io.*;
import java.util.*;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.obolibrary.robot.IOHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** @author <a href="mailto@rbca.jackson@gmail.com">Becky Jackson</a> */
public class Table {
  /** Logger. */
  private static final Logger logger = LoggerFactory.getLogger(Table.class);

  private String format = null;

  // Ordered list of Columns
  private List<Column> columns;

  // Ordered list of Rows
  private List<Row> rows;

  // Ordered list of columns to sort on
  // e.g., the first item in the list will be sorted first
  private List<Column> sortColumns;

  private RendererType displayRenderer;
  private RendererType sortRenderer = null;

  private static final Set<String> BASIC_FORMATS = Sets.newHashSet("tsv", "csv", "json", "xlsx");
  private static final String BOOTSTRAP_CSS =
      "https://stackpath.bootstrapcdn.com/bootstrap/4.5.2/css/bootstrap.min.css";
  private static final String BOOTSTRAP_JS =
      "https://stackpath.bootstrapcdn.com/bootstrap/4.5.2/js/bootstrap.min.js";

  /** Init a new Table. */
  public Table() {
    columns = new ArrayList<>();
    rows = new ArrayList<>();
    sortColumns = new ArrayList<>();
    displayRenderer = RendererType.OBJECT_RENDERER;
  }

  /**
   * Init a new Table with a format.
   *
   * @param format format of the table (tsv, csv, html)
   */
  public Table(String format) {
    this.format = format;
    columns = new ArrayList<>();
    rows = new ArrayList<>();
    sortColumns = new ArrayList<>();

    // Set renderer types based on format
    if (format == null || BASIC_FORMATS.contains(format.toLowerCase())) {
      displayRenderer = RendererType.OBJECT_RENDERER;
    } else if (format.toLowerCase().startsWith("html")) {
      displayRenderer = RendererType.OBJECT_HTML_RENDERER;
      sortRenderer = RendererType.OBJECT_RENDERER;
    } else {
      displayRenderer = RendererType.OBJECT_RENDERER;
    }
  }

  /**
   * Add a Column to the Table. This will be appended to the end of the current list of Columns.
   *
   * @param column Column to add.
   */
  public void addColumn(Column column) {
    columns.add(column);
  }

  /**
   * Add a Row to the Table. This will be appended to the end of the current list of Rows.
   *
   * @param row Row to add
   */
  public void addRow(Row row) {
    rows.add(row);
  }

  /**
   * Render the Table as a Workbook.
   *
   * @param split character to split multiple cell values on
   * @return Workbook
   */
  public Workbook asWorkbook(String split) {
    Workbook wb = new XSSFWorkbook();
    wb.createSheet();

    // Add headers
    Sheet sheet = wb.getSheetAt(0);
    org.apache.poi.ss.usermodel.Row headerRow = sheet.createRow(0);
    int colIdx = 0;
    Map<Integer, String> rules = new HashMap<>();
    for (Column c : columns) {
      String name = c.getDisplayName();
      Cell xlsxCell = headerRow.createCell(colIdx);
      xlsxCell.setCellValue(name);

      String displayRule = c.getDisplayRule();
      if (displayRule != null) {
        rules.put(colIdx, displayRule);
      }
      colIdx++;
    }

    // Maybe add rules
    if (!rules.isEmpty()) {
      org.apache.poi.ss.usermodel.Row rulesRow = sheet.createRow(sheet.getLastRowNum() + 1);
      for (int idx = 0; idx <= colIdx; idx++) {
        if (rules.containsKey(idx)) {
          String rule = rules.get(idx);
          Cell xlsxCell = rulesRow.createCell(idx);
          xlsxCell.setCellValue(rule);
        }
      }
    }

    // Add rows
    for (Row row : rows) {
      row.addToWorkbook(wb, columns, split);
    }

    // Set auto sizing
    // TODO - this takes up, relatively, a lot of time.
    /* for (int idx = 0; idx < columns.size(); idx++) {
      sheet.autoSizeColumn(idx);
    } */

    return wb;
  }

  /**
   * Get the type of display renderer
   *
   * @return RendererType
   */
  public RendererType getDisplayRendererType() {
    return displayRenderer;
  }

  /**
   * Get thee type of sort renderer
   *
   * @return RendererType
   */
  public RendererType getSortRendererType() {
    return sortRenderer;
  }

  /**
   * Get the Columns in a Table.
   *
   * @return List of Columns
   */
  public List<Column> getColumns() {
    return columns;
  }

  /**
   * Get the format of this table
   *
   * @return table format (csv, tsv, or html)
   */
  public String getFormat() {
    return format;
  }

  /**
   * Once all Columns have been added to a Table, set the columns to sort on. The columns should
   * have consecutive sort order integers which will determine the order in which they are used to
   * sort the Rows.
   */
  public void setSortColumns() {
    // Order sort columns
    Map<Integer, Column> sortColumns = new HashMap<>();
    int maxSort = 0;
    for (Column c : columns) {
      int sortOrder = c.getSortOrder();
      sortColumns.put(sortOrder, c);
      if (sortOrder > maxSort) {
        maxSort = sortOrder;
      }
    }
    int cur = 0;
    while (cur <= maxSort) {
      this.sortColumns.add(sortColumns.get(cur));
      cur++;
    }
  }

  /** Once all Rows have been added to a Table, sort Rows based on the sort Columns. */
  public void sortRows() {
    for (Column sc : sortColumns) {
      // Sort name is used to get the value
      String sortName = sc.getDisplayName();
      logger.info("Sorting on column " + sortName);
      Comparator<Row> rowComparator =
          (r1, r2) -> {
            String o1 = r1.getSortValueString(sortName);
            String o2 = r2.getSortValueString(sortName);
            if (o1.trim().isEmpty() && o2.trim().isEmpty()) return 0;
            else if (o1.trim().isEmpty()) return 1;
            else if (o2.trim().isEmpty()) return -1;
            else return o1.compareTo(o2);
          };
      if (sc.isReverseSort()) {
        rows.sort(Collections.reverseOrder(rowComparator));
      } else {
        rows.sort(rowComparator);
      }
    }
  }

  /**
   * Render the Table as List of Arrays for writing to CSV/TSV.
   *
   * @param split String to split mulitple cell values on (default |)
   * @return List of String[]
   */
  public List<String[]> toList(String split) {
    List<String[]> table = new ArrayList<>();
    String[] header = new String[columns.size()];
    Iterator<Column> iterator = columns.iterator();
    for (int i = 0; i < header.length; i++) {
      header[i] = iterator.next().getDisplayName();
    }
    table.add(header);

    for (Row row : rows) {
      String[] rowArray = row.toArray(columns, split);
      table.add(rowArray);
    }
    return table;
  }

  /**
   * Render the Table as an HTML string.
   *
   * @param split character to split multiple cell values on
   * @return HTML string
   */
  public String toHTML(String split) {
    return toHTML(split, true, false);
  }

  /**
   * Render the Table as an HTML string.
   *
   * @param split character to split multiple cell values on
   * @param standalone if true, include header
   * @return HTML string
   */
  public String toHTML(String split, boolean standalone) {
    return toHTML(split, standalone, false);
  }

  /**
   * Render the Table as an HTML string.
   *
   * @param split character to split multiple cell values on
   * @param standalone if true, include header
   * @param includeJS if true and standalone, include JS script for tooltips
   * @return HTML string
   */
  public String toHTML(String split, boolean standalone, boolean includeJS) {
    StringBuilder sb = new StringBuilder();
    if (standalone) {
      // Add opening tags, style, and maybe js scripts
      sb.append("<head>\n")
          .append("  <link rel=\"stylesheet\" href=\"")
          .append(BOOTSTRAP_CSS)
          .append("\">\n");
      if (includeJS) {
        sb.append("  <script src=\"https://code.jquery.com/jquery-3.5.1.slim.min.js\"></script>\n")
            .append(
                "  <script src=\"https://cdn.jsdelivr.net/npm/popper.js@1.16.0/dist/umd/popper.min.js\"></script>\n")
            .append("  <script src=\"")
            .append(BOOTSTRAP_JS)
            .append("\"></script>\n");
      }
      sb.append("</head>\n").append("<body>\n");
    }
    // Table start
    sb.append("<table class=\"table table-bordered table-striped\">\n")
        .append("<thead class=\"bg-dark text-white header-row\">\n")
        .append("<tr>\n");

    // Add column headers
    for (Column c : columns) {
      sb.append("  <th>").append(c.getDisplayName()).append("</th>\n");
    }
    sb.append("</tr>\n").append("</thead>\n");

    // Add all table rows
    for (Row row : rows) {
      sb.append(row.toHTML(columns, split));
    }

    sb.append("</table>\n");

    if (standalone) {
      // Add closing tag and script to activate tooltips
      sb.append("</body>\n");
      if (includeJS) {
        sb.append("<script>\n")
            .append("  $(function () {\n")
            .append("    $('[data-toggle=\"tooltip\"]').tooltip()\n")
            .append("  })")
            .append("</script>\n");
      }
    }
    return sb.toString();
  }

  public String toHTMLList() {
    StringBuilder sb = new StringBuilder();
    sb.append("<head>\n")
        .append("  <link rel=\"stylesheet\" href=\"")
        .append(BOOTSTRAP_CSS)
        .append("\">\n")
        .append("</head>\n")
        .append("<body>\n");
    for (Row row : rows) {
      sb.append(row.toHTMLList(columns));
    }
    sb.append("</body>");
    return sb.toString();
  }

  /**
   * Render the Table as a JSON string.
   *
   * @return JSON string
   */
  public String toJSON() {
    JsonArray table = new JsonArray();
    for (Row row : rows) {
      table.add(row.toJSON(columns));
    }

    Gson gson = new GsonBuilder().setPrettyPrinting().disableHtmlEscaping().create();
    return gson.toJson(table);
  }

  /**
   * Render the Table as a YAML string.
   *
   * @return YAML string
   * @throws IOException on issue writing to JSON tree
   */
  public String toYAML() throws IOException {
    JsonNode jsonNodeTree = new ObjectMapper().readTree(toJSON());
    return new YAMLMapper().writeValueAsString(jsonNodeTree);
  }

  /**
   * Given a path and a split character, write the table to a path.
   *
   * @param path path to write to
   * @param split split character for multiple cell values
   * @return true on success
   * @throws IOException on any problem opening or writing file
   */
  public boolean write(String path, String split) throws IOException {
    return write(path, split, true);
  }

  /**
   * Given a path, a split character, and boolean if standalone, write the table to a path.
   *
   * @param path path to write to
   * @param split split character for multiple cell values
   * @param standalone if true, include HTML headers for HTML output
   * @return true on success
   * @throws IOException on any problem opening or writing file
   */
  public boolean write(String path, String split, boolean standalone) throws IOException {
    File f = new File(path);
    if (format == null) {
      throw new IOException(
          String.format("Unable to write table to path '%s' with null format", path));
    }
    switch (format) {
      case "tsv":
        IOHelper.writeTable(toList(split), f, '\t');
        break;
      case "csv":
        IOHelper.writeTable(toList(split), f, ',');
        break;
      case "html":
        try (PrintWriter out = new PrintWriter(f)) {
          out.print(toHTML(split, standalone));
        }
        break;
      case "html-list":
        try (PrintWriter out = new PrintWriter(f)) {
          out.print(toHTMLList());
        }
      case "json":
        try (PrintWriter out = new PrintWriter(f)) {
          out.print(toJSON());
        }
        break;
      case "xlsx":
        try (Workbook wb = asWorkbook(split);
            FileOutputStream fos = new FileOutputStream(f)) {
          wb.write(fos);
        }
        break;
      default:
        return false;
    }
    return true;
  }
}
