package org.obolibrary.robot;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.IOException;
import java.io.StringWriter;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Set;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/** Helper methods for core tests. */
public class CoreTest {
  /** Logger. */
  private static final Logger logger = LoggerFactory.getLogger(CoreTest.class);

  /** Base IRI string for resources files. */
  protected static String base =
      "https://github.com/" + "ontodev/robot/" + "robot-core/" + "src/test/resources/";

  /** Shared data factory. */
  protected static OWLDataFactory dataFactory = OWLManager.getOWLDataFactory();

  /**
   * Load an ontology from a resource path.
   *
   * @param path the resource path for the ontology
   * @return the loaded ontology
   * @throws IOException on file problems
   */
  public OWLOntology loadOntology(String path) throws IOException {
    IOHelper ioh = new IOHelper();
    return ioh.loadOntology(this.getClass().getResourceAsStream(path));
  }

  /**
   * Load ontology using guessed catalog.
   *
   * @param path the resource path for the ontology
   * @return the loaded ontology
   * @throws IOException on file problems
   * @throws URISyntaxException on converting path to URI
   */
  public OWLOntology loadOntologyWithCatalog(String path) throws IOException, URISyntaxException {
    IOHelper ioh = new IOHelper();
    URI uri = this.getClass().getResource(path).toURI();
    return ioh.loadOntology(new File(uri));
  }

  /**
   * Load ontology using explicit catalog.
   *
   * @param ontologyIRI the IRI
   * @param catalogFile the catalog
   * @return the loaded ontology
   * @throws IOException on file problems
   * @throws OWLOntologyCreationException on creation error
   */
  public OWLOntology loadOntologyWithCatalog(IRI ontologyIRI, File catalogFile)
      throws IOException, OWLOntologyCreationException {
    // TODO: move logic here to IOHelper
    // IOHelper ioh = new IOHelper();
    OWLOntologyManager manager = OWLManager.createOWLOntologyManager();
    manager.getIRIMappers().add(new CatalogXmlIRIMapper(catalogFile));
    return manager.loadOntology(ontologyIRI);
  }

  /**
   * Given an ontology path and an ontology, assert that they are the same, and print a message to
   * the test log.
   *
   * @param leftPath the resource path for the first ontology to compare
   * @param right the second ontology to compare
   * @throws IOException on any problem
   */
  public void assertIdentical(String leftPath, OWLOntology right) throws IOException {
    OWLOntology left = loadOntology(leftPath);
    assertIdentical(left, right);
  }

  /**
   * Given two ontologies, assert that they are the same, and print a message to the test log.
   *
   * @param left the first ontology to compare
   * @param right the second ontology to compare
   * @throws IOException on any problem
   */
  public void assertIdentical(OWLOntology left, OWLOntology right) throws IOException {
    StringWriter writer = new StringWriter();
    boolean actual = DiffOperation.compare(left, right, writer);
    logger.debug(writer.toString());
    assertTrue(actual);
    assertEquals("Ontologies are identical\n", writer.toString());
  }

  /**
   * Removes all declaration axioms.
   *
   * <p>Sometimes required for tests that compare an output ontology with a previously generated
   * ontology (saving an ontology can unintentionally introduce declaration axioms)
   *
   * @param ont The target ontology
   */
  protected void removeDeclarations(OWLOntology ont) {
    Set<OWLDeclarationAxiom> decls = ont.getAxioms(AxiomType.DECLARATION);
    ont.getOWLOntologyManager().removeAxioms(ont, decls);
  }
}
